/*
 * The contents of this file are subject to the MonetDB Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.monetdb.org/Legal/MonetDBLicense
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the MonetDB Database System.
 *
 * The Initial Developer of the Original Code is CWI.
 * Portions created by CWI are Copyright (C) 1997-July 2008 CWI.
 * Copyright August 2008-2011 MonetDB B.V.
 * All Rights Reserved.
 */

#ifndef _SQL_H
#define _SQL_H

#include <sql_mem.h>

#ifdef WIN32
#ifndef LIBSQL
#define sql5_export extern __declspec(dllimport)
#else
#define sql5_export extern __declspec(dllexport)
#endif
#else
#define sql5_export extern
#endif

#include "sql_mvc.h"
#include <sql_backend.h>
#include <mal_session.h>

#include <mal_function.h>
#include <mal_stack.h>
#include <mal_interpreter.h>
#include <mal_session.h>
#include <tablet.h>
#include <streams.h>
#include <mtime.h>
#include <math.h>
#include <blob.h>
#include <mkey.h>
#include <str.h>
#include "sql_privileges.h"
#include "sql_semantic.h"
#include "sql_rel2bin.h"
#include <bin_optimizer.h>
#include "sql_decimal.h"
#include "sql_string.h"
#include "sql_qc.h"
#include "sql_env.h"
#include "sql_emptyset.h"

#include <bat/bat_storage.h>

/*
 * @-
 * The back-end structure collects the information needed to support
 * compilation and execution of the SQL code against the Monet Version 5
 * back end. Note that the back-end can be called upon by the front-end
 * to handle specific tasks, such as catalog management (sql_mvc)
 * and query execution (sql_qc). For this purpose, the front-end needs
 * access to operations defined in the back-end, in particular for
 * freeing the stack and code segment.
 * @-
 */
typedef struct backend {
	int console;
	char language;		/* 'S' or 's' or 'X' */
	mvc *mvc;
	stream *out;
	Client client;
	sql_schema *currSchema;
	sql_table *currTable;
	sql_column *currColumn;
	sql_key *currKey;
	sql_idx *currIndex;
	int 	mvc_var;	
	int	vtop;		/* top of the variable stack before the current function */
	cq *q;			/* pointer to the cached query */
} backend;

extern backend *backend_reset(backend *b);
extern backend *backend_create(mvc *m, Client c);
extern void backend_destroy(backend *b);

extern int sqlcleanup(mvc *c, int err);
extern stmt *sql_symbol2stmt(mvc *c, symbol *sym);

extern BAT *mvc_bind(mvc *m, char *sname, char *tname, char *cname, int access);
extern BAT *mvc_bind_dbat(mvc *m, char *sname, char *tname, int access);
extern BAT *mvc_bind_idxbat(mvc *m, char *sname, char *tname, char *iname, int access);

sql5_export str SQLmvc(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLtransaction(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLcommit(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLtransaction2(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLcatalog(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);

sql5_export str mvc_append_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_update_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_bind_single_wrap(int *ret, int *bid, str *name);
sql5_export str mvc_bind_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_bind_dbat_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_bind_idxbat_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_clear_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_delete_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_result_row_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_result_file_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_result_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_result_column_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_result_value_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);

sql5_export str mvc_declared_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_declared_table_column_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_drop_declared_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci );
sql5_export str mvc_drop_declared_tables_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci );

sql5_export str mvc_affected_rows_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_export_result_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_export_head_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_export_chunk_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_export_operation_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_export_value_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_import_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_bin_import_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str setVariable(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str getVariable(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str sql_variables(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_logfile(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_next_value(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_bat_next_value(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_get_value(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str mvc_diff_idxbat_wrap(bat *b, int *clientid, str *s, str *t, str *i, int *access, lng *oldversion, lng *version );
sql5_export str mvc_diff_dbat_wrap(bat *b, int *clientid, str *s, str *t, int *access, lng *oldversion, lng *version);
sql5_export str mvc_getVersion(lng *r, int *clientid);
sql5_export str mvc_restart_seq(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str zero_or_one(ptr ret, int *bid);
sql5_export str not_unique(bit *ret, int *bid);
sql5_export str not_unique_oids(bat *ret, bat *bid);
sql5_export str SQLcluster1(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLcluster2(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLshrink(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLreuse(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLvacuum(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLdrop_hash(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLnewDictionary(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLdropDictionary(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLgzcompress(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLgzdecompress(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLtruncate(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLexpand(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLoctopusBind(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLoctopusBinddbat(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLargRecord(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLrdfShred(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str month_interval_str( int *ret, str *s, int *ek, int *sk );
sql5_export str second_interval_str( lng *res, str *s, int *ek, int *sk );
sql5_export str dump_cache(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str dump_opt_stats(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str dump_trace(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str sql_storage(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str sql_rowid(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str sql_rank_grp(bat *rid, bat *bid, bat *gid, bat *gpe);
sql5_export str sql_rank(bat *rid, bat *bid);
sql5_export str sql_dense_rank_grp(bat *rid, bat *bid, bat *gid, bat *gpe);
sql5_export str sql_dense_rank(bat *rid, bat *bid);



sql5_export str bte_dec_round_wrap( bte *res, bte *v, bte *r );
sql5_export str bte_round_wrap( bte *res, bte *v, int *d, int *s, bte *r );
sql5_export str str_2dec_bte( bte *res, str *val, int *d, int *sc );
sql5_export str str_2num_bte( bte *res, str *v, int *len );
sql5_export str batstr_2dec_bte( int *res, int *val, int *d, int *sc );
sql5_export str batstr_2num_bte( int *res, int *v, int *len );




sql5_export str sht_dec_round_wrap( sht *res, sht *v, sht *r );
sql5_export str sht_round_wrap( sht *res, sht *v, int *d, int *s, bte *r );
sql5_export str str_2dec_sht( sht *res, str *val, int *d, int *sc );
sql5_export str str_2num_sht( sht *res, str *v, int *len );
sql5_export str batstr_2dec_sht( int *res, int *val, int *d, int *sc );
sql5_export str batstr_2num_sht( int *res, int *v, int *len );




sql5_export str int_dec_round_wrap( int *res, int *v, int *r );
sql5_export str int_round_wrap( int *res, int *v, int *d, int *s, bte *r );
sql5_export str str_2dec_int( int *res, str *val, int *d, int *sc );
sql5_export str str_2num_int( int *res, str *v, int *len );
sql5_export str batstr_2dec_int( int *res, int *val, int *d, int *sc );
sql5_export str batstr_2num_int( int *res, int *v, int *len );




sql5_export str wrd_dec_round_wrap( wrd *res, wrd *v, wrd *r );
sql5_export str wrd_round_wrap( wrd *res, wrd *v, int *d, int *s, bte *r );
sql5_export str str_2dec_wrd( wrd *res, str *val, int *d, int *sc );
sql5_export str str_2num_wrd( wrd *res, str *v, int *len );
sql5_export str batstr_2dec_wrd( int *res, int *val, int *d, int *sc );
sql5_export str batstr_2num_wrd( int *res, int *v, int *len );




sql5_export str lng_dec_round_wrap( lng *res, lng *v, lng *r );
sql5_export str lng_round_wrap( lng *res, lng *v, int *d, int *s, bte *r );
sql5_export str str_2dec_lng( lng *res, str *val, int *d, int *sc );
sql5_export str str_2num_lng( lng *res, str *v, int *len );
sql5_export str batstr_2dec_lng( int *res, int *val, int *d, int *sc );
sql5_export str batstr_2num_lng( int *res, int *v, int *len );






sql5_export str str_2_bit( bit *res, str *val );
sql5_export str batstr_2_bit( int *res, int *val );
sql5_export str SQLbit_2_str( str *res, bit *val );




sql5_export str str_2_oid( oid *res, str *val );
sql5_export str batstr_2_oid( int *res, int *val );
sql5_export str SQLoid_2_str( str *res, oid *val );




sql5_export str str_2_bte( bte *res, str *val );
sql5_export str batstr_2_bte( int *res, int *val );
sql5_export str SQLbte_2_str( str *res, bte *val );




sql5_export str str_2_sht( sht *res, str *val );
sql5_export str batstr_2_sht( int *res, int *val );
sql5_export str SQLsht_2_str( str *res, sht *val );




sql5_export str str_2_int( int *res, str *val );
sql5_export str batstr_2_int( int *res, int *val );
sql5_export str SQLint_2_str( str *res, int *val );




sql5_export str str_2_lng( lng *res, str *val );
sql5_export str batstr_2_lng( int *res, int *val );
sql5_export str SQLlng_2_str( str *res, lng *val );




sql5_export str str_2_wrd( wrd *res, str *val );
sql5_export str batstr_2_wrd( int *res, int *val );
sql5_export str SQLwrd_2_str( str *res, wrd *val );




sql5_export str str_2_flt( flt *res, str *val );
sql5_export str batstr_2_flt( int *res, int *val );
sql5_export str SQLflt_2_str( str *res, flt *val );




sql5_export str str_2_dbl( dbl *res, str *val );
sql5_export str batstr_2_dbl( int *res, int *val );
sql5_export str SQLdbl_2_str( str *res, dbl *val );




sql5_export str str_2_timestamp( timestamp *res, str *val );
sql5_export str batstr_2_timestamp( int *res, int *val );
sql5_export str SQLtimestamp_2_str( str *res, timestamp *val );




sql5_export str str_2_daytime( daytime *res, str *val );
sql5_export str batstr_2_daytime( int *res, int *val );
sql5_export str SQLdaytime_2_str( str *res, daytime *val );




sql5_export str str_2_date( date *res, str *val );
sql5_export str batstr_2_date( int *res, int *val );
sql5_export str SQLdate_2_str( str *res, date *val );




sql5_export str str_2_sqlblob( sqlblob* *res, str *val );
sql5_export str batstr_2_sqlblob( int *res, int *val );
sql5_export str SQLsqlblob_2_str( str *res, sqlblob *val );



sql5_export str SQLstr_cast(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
sql5_export str SQLbatstr_cast(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);



sql5_export str flt_dec_round_wrap( flt *res, flt *v, flt *r );
sql5_export str flt_round_wrap( flt *res, flt *v, bte *r );
sql5_export str flt_trunc_wrap( flt *res, flt *v, int *r );




sql5_export str dbl_dec_round_wrap( dbl *res, dbl *v, dbl *r );
sql5_export str dbl_round_wrap( dbl *res, dbl *v, bte *r );
sql5_export str dbl_trunc_wrap( dbl *res, dbl *v, int *r );



#define radians(x)       ((x) * 3.14159265358979323846 /180.0 )
#define degrees(x)       ((x) * 180.0/3.14159265358979323846 )
sql5_export str SQLdbl_alpha(dbl *res, dbl *decl, dbl *theta);
sql5_export str SQLbat_alpha(bat *res, bat *decl, dbl *theta);



sql5_export str month_interval_bte( int *ret, bte *s, int *ek, int *sk );
sql5_export str second_interval_bte( lng *res, bte *s, int *ek, int *sk );




sql5_export str month_interval_sht( int *ret, sht *s, int *ek, int *sk );
sql5_export str second_interval_sht( lng *res, sht *s, int *ek, int *sk );




sql5_export str month_interval_int( int *ret, int *s, int *ek, int *sk );
sql5_export str second_interval_int( lng *res, int *s, int *ek, int *sk );




sql5_export str month_interval_wrd( int *ret, wrd *s, int *ek, int *sk );
sql5_export str second_interval_wrd( lng *res, wrd *s, int *ek, int *sk );




sql5_export str month_interval_lng( int *ret, lng *s, int *ek, int *sk );
sql5_export str second_interval_lng( lng *res, lng *s, int *ek, int *sk );




sql5_export str month_interval_daytime( int *ret, daytime *s, int *ek, int *sk );
sql5_export str second_interval_daytime( lng *res, daytime *s, int *ek, int *sk );







sql5_export str bte_2_bte( bte *res, bte *v );
sql5_export str batbte_2_bte( int *res, int *v );



sql5_export str bte_dec2_bte( bte *res, int *s1, bte *v );
sql5_export str bte_dec2dec_bte( bte *res, int *S1, bte *v, int *d2, int *S2 );
sql5_export str bte_num2dec_bte( bte *res, bte *v, int *d2, int *s2 );
sql5_export str batbte_dec2_bte( int *res, int *s1, int *v );
sql5_export str batbte_dec2dec_bte( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batbte_num2dec_bte( int *res, int *v, int *d2, int *s2 );





sql5_export str bte_2_sht( sht *res, bte *v );
sql5_export str batbte_2_sht( int *res, int *v );



sql5_export str bte_dec2_sht( sht *res, int *s1, bte *v );
sql5_export str bte_dec2dec_sht( sht *res, int *S1, bte *v, int *d2, int *S2 );
sql5_export str bte_num2dec_sht( sht *res, bte *v, int *d2, int *s2 );
sql5_export str batbte_dec2_sht( int *res, int *s1, int *v );
sql5_export str batbte_dec2dec_sht( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batbte_num2dec_sht( int *res, int *v, int *d2, int *s2 );





sql5_export str sht_2_sht( sht *res, sht *v );
sql5_export str batsht_2_sht( int *res, int *v );



sql5_export str sht_dec2_sht( sht *res, int *s1, sht *v );
sql5_export str sht_dec2dec_sht( sht *res, int *S1, sht *v, int *d2, int *S2 );
sql5_export str sht_num2dec_sht( sht *res, sht *v, int *d2, int *s2 );
sql5_export str batsht_dec2_sht( int *res, int *s1, int *v );
sql5_export str batsht_dec2dec_sht( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batsht_num2dec_sht( int *res, int *v, int *d2, int *s2 );





sql5_export str bte_2_int( int *res, bte *v );
sql5_export str batbte_2_int( int *res, int *v );



sql5_export str bte_dec2_int( int *res, int *s1, bte *v );
sql5_export str bte_dec2dec_int( int *res, int *S1, bte *v, int *d2, int *S2 );
sql5_export str bte_num2dec_int( int *res, bte *v, int *d2, int *s2 );
sql5_export str batbte_dec2_int( int *res, int *s1, int *v );
sql5_export str batbte_dec2dec_int( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batbte_num2dec_int( int *res, int *v, int *d2, int *s2 );





sql5_export str sht_2_int( int *res, sht *v );
sql5_export str batsht_2_int( int *res, int *v );



sql5_export str sht_dec2_int( int *res, int *s1, sht *v );
sql5_export str sht_dec2dec_int( int *res, int *S1, sht *v, int *d2, int *S2 );
sql5_export str sht_num2dec_int( int *res, sht *v, int *d2, int *s2 );
sql5_export str batsht_dec2_int( int *res, int *s1, int *v );
sql5_export str batsht_dec2dec_int( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batsht_num2dec_int( int *res, int *v, int *d2, int *s2 );





sql5_export str int_2_int( int *res, int *v );
sql5_export str batint_2_int( int *res, int *v );



sql5_export str int_dec2_int( int *res, int *s1, int *v );
sql5_export str int_dec2dec_int( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str int_num2dec_int( int *res, int *v, int *d2, int *s2 );
sql5_export str batint_dec2_int( int *res, int *s1, int *v );
sql5_export str batint_dec2dec_int( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batint_num2dec_int( int *res, int *v, int *d2, int *s2 );





sql5_export str bte_2_wrd( wrd *res, bte *v );
sql5_export str batbte_2_wrd( int *res, int *v );



sql5_export str bte_dec2_wrd( wrd *res, int *s1, bte *v );
sql5_export str bte_dec2dec_wrd( wrd *res, int *S1, bte *v, int *d2, int *S2 );
sql5_export str bte_num2dec_wrd( wrd *res, bte *v, int *d2, int *s2 );
sql5_export str batbte_dec2_wrd( int *res, int *s1, int *v );
sql5_export str batbte_dec2dec_wrd( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batbte_num2dec_wrd( int *res, int *v, int *d2, int *s2 );





sql5_export str sht_2_wrd( wrd *res, sht *v );
sql5_export str batsht_2_wrd( int *res, int *v );



sql5_export str sht_dec2_wrd( wrd *res, int *s1, sht *v );
sql5_export str sht_dec2dec_wrd( wrd *res, int *S1, sht *v, int *d2, int *S2 );
sql5_export str sht_num2dec_wrd( wrd *res, sht *v, int *d2, int *s2 );
sql5_export str batsht_dec2_wrd( int *res, int *s1, int *v );
sql5_export str batsht_dec2dec_wrd( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batsht_num2dec_wrd( int *res, int *v, int *d2, int *s2 );





sql5_export str int_2_wrd( wrd *res, int *v );
sql5_export str batint_2_wrd( int *res, int *v );



sql5_export str int_dec2_wrd( wrd *res, int *s1, int *v );
sql5_export str int_dec2dec_wrd( wrd *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str int_num2dec_wrd( wrd *res, int *v, int *d2, int *s2 );
sql5_export str batint_dec2_wrd( int *res, int *s1, int *v );
sql5_export str batint_dec2dec_wrd( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batint_num2dec_wrd( int *res, int *v, int *d2, int *s2 );





sql5_export str wrd_2_wrd( wrd *res, wrd *v );
sql5_export str batwrd_2_wrd( int *res, int *v );



sql5_export str wrd_dec2_wrd( wrd *res, int *s1, wrd *v );
sql5_export str wrd_dec2dec_wrd( wrd *res, int *S1, wrd *v, int *d2, int *S2 );
sql5_export str wrd_num2dec_wrd( wrd *res, wrd *v, int *d2, int *s2 );
sql5_export str batwrd_dec2_wrd( int *res, int *s1, int *v );
sql5_export str batwrd_dec2dec_wrd( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batwrd_num2dec_wrd( int *res, int *v, int *d2, int *s2 );





sql5_export str bte_2_lng( lng *res, bte *v );
sql5_export str batbte_2_lng( int *res, int *v );



sql5_export str bte_dec2_lng( lng *res, int *s1, bte *v );
sql5_export str bte_dec2dec_lng( lng *res, int *S1, bte *v, int *d2, int *S2 );
sql5_export str bte_num2dec_lng( lng *res, bte *v, int *d2, int *s2 );
sql5_export str batbte_dec2_lng( int *res, int *s1, int *v );
sql5_export str batbte_dec2dec_lng( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batbte_num2dec_lng( int *res, int *v, int *d2, int *s2 );





sql5_export str sht_2_lng( lng *res, sht *v );
sql5_export str batsht_2_lng( int *res, int *v );



sql5_export str sht_dec2_lng( lng *res, int *s1, sht *v );
sql5_export str sht_dec2dec_lng( lng *res, int *S1, sht *v, int *d2, int *S2 );
sql5_export str sht_num2dec_lng( lng *res, sht *v, int *d2, int *s2 );
sql5_export str batsht_dec2_lng( int *res, int *s1, int *v );
sql5_export str batsht_dec2dec_lng( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batsht_num2dec_lng( int *res, int *v, int *d2, int *s2 );





sql5_export str int_2_lng( lng *res, int *v );
sql5_export str batint_2_lng( int *res, int *v );



sql5_export str int_dec2_lng( lng *res, int *s1, int *v );
sql5_export str int_dec2dec_lng( lng *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str int_num2dec_lng( lng *res, int *v, int *d2, int *s2 );
sql5_export str batint_dec2_lng( int *res, int *s1, int *v );
sql5_export str batint_dec2dec_lng( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batint_num2dec_lng( int *res, int *v, int *d2, int *s2 );





sql5_export str wrd_2_lng( lng *res, wrd *v );
sql5_export str batwrd_2_lng( int *res, int *v );



sql5_export str wrd_dec2_lng( lng *res, int *s1, wrd *v );
sql5_export str wrd_dec2dec_lng( lng *res, int *S1, wrd *v, int *d2, int *S2 );
sql5_export str wrd_num2dec_lng( lng *res, wrd *v, int *d2, int *s2 );
sql5_export str batwrd_dec2_lng( int *res, int *s1, int *v );
sql5_export str batwrd_dec2dec_lng( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batwrd_num2dec_lng( int *res, int *v, int *d2, int *s2 );





sql5_export str lng_2_lng( lng *res, lng *v );
sql5_export str batlng_2_lng( int *res, int *v );



sql5_export str lng_dec2_lng( lng *res, int *s1, lng *v );
sql5_export str lng_dec2dec_lng( lng *res, int *S1, lng *v, int *d2, int *S2 );
sql5_export str lng_num2dec_lng( lng *res, lng *v, int *d2, int *s2 );
sql5_export str batlng_dec2_lng( int *res, int *s1, int *v );
sql5_export str batlng_dec2dec_lng( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batlng_num2dec_lng( int *res, int *v, int *d2, int *s2 );







sql5_export str flt_2_bte( bte *res, flt *v );
sql5_export str batflt_2_bte( int *res, int *v );



sql5_export str flt_num2dec_bte( bte *res, flt *v, int *d2, int *s2 );
sql5_export str batflt_num2dec_bte( int *res, int *v, int *d2, int *s2 );





sql5_export str flt_2_sht( sht *res, flt *v );
sql5_export str batflt_2_sht( int *res, int *v );



sql5_export str flt_num2dec_sht( sht *res, flt *v, int *d2, int *s2 );
sql5_export str batflt_num2dec_sht( int *res, int *v, int *d2, int *s2 );





sql5_export str flt_2_int( int *res, flt *v );
sql5_export str batflt_2_int( int *res, int *v );



sql5_export str flt_num2dec_int( int *res, flt *v, int *d2, int *s2 );
sql5_export str batflt_num2dec_int( int *res, int *v, int *d2, int *s2 );





sql5_export str flt_2_wrd( wrd *res, flt *v );
sql5_export str batflt_2_wrd( int *res, int *v );



sql5_export str flt_num2dec_wrd( wrd *res, flt *v, int *d2, int *s2 );
sql5_export str batflt_num2dec_wrd( int *res, int *v, int *d2, int *s2 );





sql5_export str flt_2_lng( lng *res, flt *v );
sql5_export str batflt_2_lng( int *res, int *v );



sql5_export str flt_num2dec_lng( lng *res, flt *v, int *d2, int *s2 );
sql5_export str batflt_num2dec_lng( int *res, int *v, int *d2, int *s2 );





sql5_export str dbl_2_bte( bte *res, dbl *v );
sql5_export str batdbl_2_bte( int *res, int *v );



sql5_export str dbl_num2dec_bte( bte *res, dbl *v, int *d2, int *s2 );
sql5_export str batdbl_num2dec_bte( int *res, int *v, int *d2, int *s2 );





sql5_export str dbl_2_sht( sht *res, dbl *v );
sql5_export str batdbl_2_sht( int *res, int *v );



sql5_export str dbl_num2dec_sht( sht *res, dbl *v, int *d2, int *s2 );
sql5_export str batdbl_num2dec_sht( int *res, int *v, int *d2, int *s2 );





sql5_export str dbl_2_int( int *res, dbl *v );
sql5_export str batdbl_2_int( int *res, int *v );



sql5_export str dbl_num2dec_int( int *res, dbl *v, int *d2, int *s2 );
sql5_export str batdbl_num2dec_int( int *res, int *v, int *d2, int *s2 );





sql5_export str dbl_2_wrd( wrd *res, dbl *v );
sql5_export str batdbl_2_wrd( int *res, int *v );



sql5_export str dbl_num2dec_wrd( wrd *res, dbl *v, int *d2, int *s2 );
sql5_export str batdbl_num2dec_wrd( int *res, int *v, int *d2, int *s2 );





sql5_export str dbl_2_lng( lng *res, dbl *v );
sql5_export str batdbl_2_lng( int *res, int *v );



sql5_export str dbl_num2dec_lng( lng *res, dbl *v, int *d2, int *s2 );
sql5_export str batdbl_num2dec_lng( int *res, int *v, int *d2, int *s2 );







sql5_export str bte_2_flt( flt *res, bte *v );
sql5_export str batbte_2_flt( int *res, int *v );



sql5_export str bte_dec2_flt( flt *res, int *s1, bte *v );
sql5_export str bte_dec2dec_flt( flt *res, int *S1, bte *v, int *d2, int *S2 );
sql5_export str bte_num2dec_flt( flt *res, bte *v, int *d2, int *s2 );
sql5_export str batbte_dec2_flt( int *res, int *s1, int *v );
sql5_export str batbte_dec2dec_flt( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batbte_num2dec_flt( int *res, int *v, int *d2, int *s2 );





sql5_export str sht_2_flt( flt *res, sht *v );
sql5_export str batsht_2_flt( int *res, int *v );



sql5_export str sht_dec2_flt( flt *res, int *s1, sht *v );
sql5_export str sht_dec2dec_flt( flt *res, int *S1, sht *v, int *d2, int *S2 );
sql5_export str sht_num2dec_flt( flt *res, sht *v, int *d2, int *s2 );
sql5_export str batsht_dec2_flt( int *res, int *s1, int *v );
sql5_export str batsht_dec2dec_flt( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batsht_num2dec_flt( int *res, int *v, int *d2, int *s2 );





sql5_export str int_2_flt( flt *res, int *v );
sql5_export str batint_2_flt( int *res, int *v );



sql5_export str int_dec2_flt( flt *res, int *s1, int *v );
sql5_export str int_dec2dec_flt( flt *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str int_num2dec_flt( flt *res, int *v, int *d2, int *s2 );
sql5_export str batint_dec2_flt( int *res, int *s1, int *v );
sql5_export str batint_dec2dec_flt( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batint_num2dec_flt( int *res, int *v, int *d2, int *s2 );





sql5_export str wrd_2_flt( flt *res, wrd *v );
sql5_export str batwrd_2_flt( int *res, int *v );



sql5_export str wrd_dec2_flt( flt *res, int *s1, wrd *v );
sql5_export str wrd_dec2dec_flt( flt *res, int *S1, wrd *v, int *d2, int *S2 );
sql5_export str wrd_num2dec_flt( flt *res, wrd *v, int *d2, int *s2 );
sql5_export str batwrd_dec2_flt( int *res, int *s1, int *v );
sql5_export str batwrd_dec2dec_flt( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batwrd_num2dec_flt( int *res, int *v, int *d2, int *s2 );





sql5_export str lng_2_flt( flt *res, lng *v );
sql5_export str batlng_2_flt( int *res, int *v );



sql5_export str lng_dec2_flt( flt *res, int *s1, lng *v );
sql5_export str lng_dec2dec_flt( flt *res, int *S1, lng *v, int *d2, int *S2 );
sql5_export str lng_num2dec_flt( flt *res, lng *v, int *d2, int *s2 );
sql5_export str batlng_dec2_flt( int *res, int *s1, int *v );
sql5_export str batlng_dec2dec_flt( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batlng_num2dec_flt( int *res, int *v, int *d2, int *s2 );





sql5_export str bte_2_dbl( dbl *res, bte *v );
sql5_export str batbte_2_dbl( int *res, int *v );



sql5_export str bte_dec2_dbl( dbl *res, int *s1, bte *v );
sql5_export str bte_dec2dec_dbl( dbl *res, int *S1, bte *v, int *d2, int *S2 );
sql5_export str bte_num2dec_dbl( dbl *res, bte *v, int *d2, int *s2 );
sql5_export str batbte_dec2_dbl( int *res, int *s1, int *v );
sql5_export str batbte_dec2dec_dbl( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batbte_num2dec_dbl( int *res, int *v, int *d2, int *s2 );





sql5_export str sht_2_dbl( dbl *res, sht *v );
sql5_export str batsht_2_dbl( int *res, int *v );



sql5_export str sht_dec2_dbl( dbl *res, int *s1, sht *v );
sql5_export str sht_dec2dec_dbl( dbl *res, int *S1, sht *v, int *d2, int *S2 );
sql5_export str sht_num2dec_dbl( dbl *res, sht *v, int *d2, int *s2 );
sql5_export str batsht_dec2_dbl( int *res, int *s1, int *v );
sql5_export str batsht_dec2dec_dbl( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batsht_num2dec_dbl( int *res, int *v, int *d2, int *s2 );





sql5_export str int_2_dbl( dbl *res, int *v );
sql5_export str batint_2_dbl( int *res, int *v );



sql5_export str int_dec2_dbl( dbl *res, int *s1, int *v );
sql5_export str int_dec2dec_dbl( dbl *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str int_num2dec_dbl( dbl *res, int *v, int *d2, int *s2 );
sql5_export str batint_dec2_dbl( int *res, int *s1, int *v );
sql5_export str batint_dec2dec_dbl( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batint_num2dec_dbl( int *res, int *v, int *d2, int *s2 );





sql5_export str wrd_2_dbl( dbl *res, wrd *v );
sql5_export str batwrd_2_dbl( int *res, int *v );



sql5_export str wrd_dec2_dbl( dbl *res, int *s1, wrd *v );
sql5_export str wrd_dec2dec_dbl( dbl *res, int *S1, wrd *v, int *d2, int *S2 );
sql5_export str wrd_num2dec_dbl( dbl *res, wrd *v, int *d2, int *s2 );
sql5_export str batwrd_dec2_dbl( int *res, int *s1, int *v );
sql5_export str batwrd_dec2dec_dbl( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batwrd_num2dec_dbl( int *res, int *v, int *d2, int *s2 );





sql5_export str lng_2_dbl( dbl *res, lng *v );
sql5_export str batlng_2_dbl( int *res, int *v );



sql5_export str lng_dec2_dbl( dbl *res, int *s1, lng *v );
sql5_export str lng_dec2dec_dbl( dbl *res, int *S1, lng *v, int *d2, int *S2 );
sql5_export str lng_num2dec_dbl( dbl *res, lng *v, int *d2, int *s2 );
sql5_export str batlng_dec2_dbl( int *res, int *s1, int *v );
sql5_export str batlng_dec2dec_dbl( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batlng_num2dec_dbl( int *res, int *v, int *d2, int *s2 );







sql5_export str sht_2_bte( bte *res, sht *v );
sql5_export str batsht_2_bte( int *res, int *v );



sql5_export str sht_dec2_bte( bte *res, int *s1, sht *v );
sql5_export str sht_dec2dec_bte( bte *res, int *S1, sht *v, int *d2, int *S2 );
sql5_export str sht_num2dec_bte( bte *res, sht *v, int *d2, int *s2 );
sql5_export str batsht_dec2_bte( int *res, int *s1, int *v );
sql5_export str batsht_dec2dec_bte( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batsht_num2dec_bte( int *res, int *v, int *d2, int *s2 );





sql5_export str int_2_bte( bte *res, int *v );
sql5_export str batint_2_bte( int *res, int *v );



sql5_export str int_dec2_bte( bte *res, int *s1, int *v );
sql5_export str int_dec2dec_bte( bte *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str int_num2dec_bte( bte *res, int *v, int *d2, int *s2 );
sql5_export str batint_dec2_bte( int *res, int *s1, int *v );
sql5_export str batint_dec2dec_bte( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batint_num2dec_bte( int *res, int *v, int *d2, int *s2 );





sql5_export str wrd_2_bte( bte *res, wrd *v );
sql5_export str batwrd_2_bte( int *res, int *v );



sql5_export str wrd_dec2_bte( bte *res, int *s1, wrd *v );
sql5_export str wrd_dec2dec_bte( bte *res, int *S1, wrd *v, int *d2, int *S2 );
sql5_export str wrd_num2dec_bte( bte *res, wrd *v, int *d2, int *s2 );
sql5_export str batwrd_dec2_bte( int *res, int *s1, int *v );
sql5_export str batwrd_dec2dec_bte( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batwrd_num2dec_bte( int *res, int *v, int *d2, int *s2 );





sql5_export str lng_2_bte( bte *res, lng *v );
sql5_export str batlng_2_bte( int *res, int *v );



sql5_export str lng_dec2_bte( bte *res, int *s1, lng *v );
sql5_export str lng_dec2dec_bte( bte *res, int *S1, lng *v, int *d2, int *S2 );
sql5_export str lng_num2dec_bte( bte *res, lng *v, int *d2, int *s2 );
sql5_export str batlng_dec2_bte( int *res, int *s1, int *v );
sql5_export str batlng_dec2dec_bte( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batlng_num2dec_bte( int *res, int *v, int *d2, int *s2 );





sql5_export str int_2_sht( sht *res, int *v );
sql5_export str batint_2_sht( int *res, int *v );



sql5_export str int_dec2_sht( sht *res, int *s1, int *v );
sql5_export str int_dec2dec_sht( sht *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str int_num2dec_sht( sht *res, int *v, int *d2, int *s2 );
sql5_export str batint_dec2_sht( int *res, int *s1, int *v );
sql5_export str batint_dec2dec_sht( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batint_num2dec_sht( int *res, int *v, int *d2, int *s2 );





sql5_export str wrd_2_sht( sht *res, wrd *v );
sql5_export str batwrd_2_sht( int *res, int *v );



sql5_export str wrd_dec2_sht( sht *res, int *s1, wrd *v );
sql5_export str wrd_dec2dec_sht( sht *res, int *S1, wrd *v, int *d2, int *S2 );
sql5_export str wrd_num2dec_sht( sht *res, wrd *v, int *d2, int *s2 );
sql5_export str batwrd_dec2_sht( int *res, int *s1, int *v );
sql5_export str batwrd_dec2dec_sht( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batwrd_num2dec_sht( int *res, int *v, int *d2, int *s2 );





sql5_export str lng_2_sht( sht *res, lng *v );
sql5_export str batlng_2_sht( int *res, int *v );



sql5_export str lng_dec2_sht( sht *res, int *s1, lng *v );
sql5_export str lng_dec2dec_sht( sht *res, int *S1, lng *v, int *d2, int *S2 );
sql5_export str lng_num2dec_sht( sht *res, lng *v, int *d2, int *s2 );
sql5_export str batlng_dec2_sht( int *res, int *s1, int *v );
sql5_export str batlng_dec2dec_sht( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batlng_num2dec_sht( int *res, int *v, int *d2, int *s2 );





sql5_export str wrd_2_int( int *res, wrd *v );
sql5_export str batwrd_2_int( int *res, int *v );



sql5_export str wrd_dec2_int( int *res, int *s1, wrd *v );
sql5_export str wrd_dec2dec_int( int *res, int *S1, wrd *v, int *d2, int *S2 );
sql5_export str wrd_num2dec_int( int *res, wrd *v, int *d2, int *s2 );
sql5_export str batwrd_dec2_int( int *res, int *s1, int *v );
sql5_export str batwrd_dec2dec_int( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batwrd_num2dec_int( int *res, int *v, int *d2, int *s2 );





sql5_export str lng_2_int( int *res, lng *v );
sql5_export str batlng_2_int( int *res, int *v );



sql5_export str lng_dec2_int( int *res, int *s1, lng *v );
sql5_export str lng_dec2dec_int( int *res, int *S1, lng *v, int *d2, int *S2 );
sql5_export str lng_num2dec_int( int *res, lng *v, int *d2, int *s2 );
sql5_export str batlng_dec2_int( int *res, int *s1, int *v );
sql5_export str batlng_dec2dec_int( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batlng_num2dec_int( int *res, int *v, int *d2, int *s2 );





sql5_export str lng_2_wrd( wrd *res, lng *v );
sql5_export str batlng_2_wrd( int *res, int *v );



sql5_export str lng_dec2_wrd( wrd *res, int *s1, lng *v );
sql5_export str lng_dec2dec_wrd( wrd *res, int *S1, lng *v, int *d2, int *S2 );
sql5_export str lng_num2dec_wrd( wrd *res, lng *v, int *d2, int *s2 );
sql5_export str batlng_dec2_wrd( int *res, int *s1, int *v );
sql5_export str batlng_dec2dec_wrd( int *res, int *S1, int *v, int *d2, int *S2 );
sql5_export str batlng_num2dec_wrd( int *res, int *v, int *d2, int *s2 );




sql5_export str getContext(Client cntxt, MalBlkPtr mb, mvc **c, backend **b );

sql5_export void freeVariables(Client c, MalBlkPtr mb, MalStkPtr glb, int start);

#endif /* _SQL_H */



sql5_export str second_interval_daytime( lng *res, daytime *s, int *ek, int *sk );

